# -*- coding: utf-8 -*-

""" arch-wiki-search (c) Clem Lorteau 2025
License: MIT
"""

__version__ = '20250825'
__name__ = 'arch_wiki_search'
__author__ = 'Clem Lorteau'
__license__ = 'MIT'

__url__ = 'https://github.com/clorteau/arch-wiki-search'
__newwikirequesturl__ = 'https://#TODO'

import logging

class CustomFormatter(logging.Formatter):
    grey = '\x1b[38;20m'
    yellow = '\x1b[33;20m'
    green = '\033[32m'
    red = '\x1b[31;20m'
    bold_red = '\x1b[31;1m'
    reset = '\x1b[0m'
    # format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)'
    fields = ' %(message)s'

    FORMATS = {
        logging.DEBUG: grey + fields + reset,
        logging.INFO: green + '🡪' + fields + reset,
        logging.WARNING: yellow + '⚠' + fields + reset,
        logging.ERROR: red + '✖' + fields + reset,
        logging.CRITICAL: bold_red + '✖✖' + fields + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

class AIOHTTPCustomFormatter(CustomFormatter):
    fields = '%a %t "%r" %s %b "%{Referer}i" "%{User-Agent}i"'


logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)




# -*- coding: utf-8 -*-

""" arch-wiki-search (c) Clem Lorteau 2025
License: MIT
"""

import os
import sys
import logging
import asyncio
import traceback
from datetime import timedelta
from aiohttp import web, DummyCookieJar
from concurrent.futures import ThreadPoolExecutor
from aiohttp_client_cache import CachedSession, FileBackend

try:
    import converters
    from __init__ import logger
except ModuleNotFoundError:
    from arch_wiki_search import converters, logger

class CachingProxy:
    """Asynchronous caching http proxy that caches for a long time, manipulates responses,
    and only serves one top domain
    """
    base_url = ''
    cache_dir = ''
    expire_days = 8
    cache = None
    app = None
    debug = False
    port = 8888 #TODO: set to available port automatically
    runner = None

    def _hsize(self, size):
        for unit in ['B', 'KB', 'MB', 'GB']:
            if size < 1024:
                return f"{size:.2f}{unit}"
            size /= 1024

    async def _printcachesize(self):
        def _dirsize(path):
            size = 0
            for dirpath, dirnames, filenames in os.walk(path):
                for f in filenames:
                    fp = os.path.join(dirpath, f)
                    size += os.path.getsize(fp)
            return size
        # run above blocking code asynchronously on executor
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor() as executor:
            result = await loop.run_in_executor(executor, _dirsize, self.cache_dir)
            logger.info(f'Cache size: {self._hsize(result)}')

    async def printcachesize(self):
        """Asynchronously calculate total cache size and output in human readable format
        """
        assert self.cache_dir != ''
        await self._printcachesize()
    
    async def start(self):
        assert self.cache != None            
        server = web.Server(self._get_handler, debug=self.debug)

        #start in separate thread
        self.runner = web.ServerRunner(server)
        if (self.debug): logging.basicConfig(level=logging.DEBUG)
        await self.runner.setup()
        site = web.TCPSite(self.runner, 'localhost', self.port)
        await site.start()
    
        logger.info(f'Serving wiki on http://localhost:{self.port} - press < ctrl-c > to stop')

    async def stop(self):
        await self.runner.cleanup()

    async def fetch(self, urlpath):
        """Retrieves contents at base_url/urlpath
        """
        #TODO: pre-cache one-level of links
        url = self.base_url + '/' + urlpath
        resp = None
        ignore_cookies = DummyCookieJar()
        async with CachedSession(cache=self.cache, cookie_jar=ignore_cookies) as session:
            try:
                resp = await session.get(f'{url}')
            except Exception as e:
                msg = f'Failed to fetch URL: {url}'
                trace = traceback.format_exc()
                logger.error(f'{msg}\n{e}')
                text = f'\
                    <!DOCTYPE html><html>\
                    <h3>{msg}</h3>'
                if (self.debug): text += f'<code>{trace.replace('\n', '<br/>\n')}</code>'
                text += '</html>'
                return web.Response(content_type='text/html', text=text)
        assert resp != None
        expires = resp.expires.isoformat() if resp.expires else 'Never'
        logger.debug(f'{resp.url} expires: {expires}')
        
        return resp

    async def _get_handler(self, request, ):
        """Fetches the requested page, manipulates it and responds with it
        """
        logger.debug(f'Got request: {request}')

        #the full URL to fetch is passed as the request's path; extract the target path
        url = request.raw_path
        url = url.lstrip('/')
        path = url.replace(self.base_url, '')

        response = await self.fetch(path)
        if self.conv == 'raw':
            converter = converters.RawConverter(response, self.base_url, self.port)
        elif self.conv == 'clean':
            converter = converters.CleanHTMLConverter(response, self.base_url, self.port)
        elif self.conv == 'txt':
            converter = converters.TxtConverter(response, self.base_url, self.port)
        else:
            #TODO: detect if running in graphical envrionment or console
            converter = converters.RawConverter(response, self.base_url, self.port)
        newresponse = await converter.convert()
        await newresponse.prepare(request)
        return newresponse

    def __init__(self, base_url, cache_dir='', expire_days=30, debug=False, conv=''):
        self.base_url = base_url
        self.expire_days = expire_days
        self.cache_dir = cache_dir
        self.debug = debug
        self.conv = conv

        if (not self.base_url.startswith(('http://', 'https://'))):
            err = f'Unsupported url: {self.base_url}'
            logger.error(err)
            sys.exit(-2)

        if os.name == 'posix': 
            self.cache_dir = os.path.join(os.path.expanduser('~'), '.cache', __name__)
        elif os.name == 'nt': 
            self.cache_dir = os.path.join(os.path.expanduser('~'), 'AppData', 'Local', __name__)

        if os.path.isdir(self.cache_dir):
            if os.access(self.cache_dir, os.W_OK):
                logger.debug(f'The cache directory {self.cache_dir} exists and is writable')
            else:
                err = f'The cache directory {self.cache_dir} is not writable'
                logger.critical(err)
                print(traceback.format_exc())
                sys.exit(-4)
        else:
            try:
                os.makedirs(self.cache_dir)
                logger.info(f'Created cache directory {self.cache_dir}')
            except Exception as e:
                logger.critical(f'Failed to create cache directory {self.cache_dir}')
                print(traceback.format_exc())
                sys.exit(-4)

        self.cache = FileBackend(
            cache_name = self.cache_dir,
            expire_after = timedelta(days=self.expire_days),
            #only cache these responses
            allowed_codes = (200, #ok
                             301, #permanent move
                             308, #permanent redirect
                            ),
        )

        



# -*- coding: utf-8 -*-

""" arch-wiki-search (c) Clem Lorteau 2025
License: MIT
"""

import os
import yaml
try:
    from __init__ import __name__, logger
except ModuleNotFoundError:
    from arch_wiki_search.arch_wiki_search import __name__, logger
    
class Wiki:
    name = ''
    url = ''
    searchstring = ''

    def __init__(self, name, url, searchstring):
        self.name = name
        self.url = url
        self.searchstring = searchstring

class Wikis(set):
    filename = ''
    dirs = []

    def getnames(self):
        names = []
        for i, w in enumerate(self):
            names.append(w.name)
        return sorted(names)

    def __init__(self, filename='wikis.yaml'):
        self.filename = filename
        super().__init__()
        self.dirs.append(os.path.dirname(os.path.realpath(__file__)))
        if os.name == 'posix': 
            configdir = os.path.join(os.path.expanduser('~'), '.config', __name__)
        elif os.name == 'nt': 
            configdir = os.path.join(os.path.expanduser('~'), 'AppData', 'Local', __name__)
        self.dirs.append(configdir)
        for d in self.dirs:
            path = d + '/' + self.filename
            try:
                f = open(path, 'r')
                docs = yaml.safe_load_all(f)
                for doc in docs:
                    self.add(Wiki(doc['name'], doc['url'], doc['searchstring']))
                f.close()
            except Exception as e:
                logger.debug(f'Could not load known wikis file {path}: {e}')
        if len(self) == 0:
            logger.warning('No known wikis found')
        else:
            logger.debug('Known wikis: ' + str(self))
        




""" arch-wiki-search (c) Clem Lorteau 2025
License: MIT
"""

import os
from datetime import datetime
from zipfile import ZipFile, ZIP_DEFLATED

# try:
#     from __init__ import __name__, logger
# except ModuleNotFoundError:
#     from arch_wiki_search.arch_wiki_search import __name__

class ZIP:
    def __init__(self):
        self.timestamp = '{:%Y%m%d_%H-%M-%S}'.format(datetime.now())
    
    def export(self, dir_path, out_path='.'):
        file_name = f'{out_path}/{__name__}-{self.timestamp}.zip'
        # if not dir_path.endswith('/'):
        #     dir_path += '/' #don't create a top level directory
        try:
            with ZipFile(file_name, 'w', ZIP_DEFLATED) as zfile:
                for root, dirs, files in os.walk(dir_path):
                    for file in files:
                        zfile.write(os.path.join(root, file),
                                    os.path.relpath(os.path.join(root, file),
                                                    os.path.join(dir_path, '..')))
                logger.info(f'Export from \'{dir_path}\' to \'{file_name}\' successful')
        except Exception as e:
            msg = f'Failed creating export \'{file_name}\' from \'{dir_path}\':\n{e}'
            logger.critical(msg)
            raise e

    def merge(self, dir_path, inzip):
        try:
            with ZipFile(inzip, 'r') as zfile:
                zfile.extractall(dir_path) #TODO: validate import
            logger.info(f'Import from {inzip} to {dir_path} successful')
        except Exception as e:
            msg = f'Failed import to {dir_path} from {inzip}:\n{e}'
            logger.critical(msg)


# -*- coding: utf-8 -*-

""" arch-wiki-search (c) Clem Lorteau 2025
License: MIT
"""

import warnings
import html5lib
import lxml_html_clean
from aiohttp import web
from aiohttp_client_cache import CachedResponse
from bs4 import BeautifulSoup, XMLParsedAsHTMLWarning

try:
    from __init__ import logger
except ModuleNotFoundError:
    from arch_wiki_search import logger
    
class RawConverter:
    """Manipulates a aiohttp.ClientResponse to convert contents
    TODO: only convert if original response status is 200 ok, otherwise return an error page
    """
    def _links_to_local(self):
        """Rewrite links by appending them to our local proxy
        """
        return self.text.replace(self.base_url, f'http://localhost:{self.port}')

    def __init__(self, response, base_url, port):
        self.base_url = base_url
        self.port = port
        newresponse = web.Response(status=response.status, content_type=response.content_type)
        self.response = response  
        self.newresponse = newresponse

    async def convert(self):
        try:
            self.text = await self.response.text()
        except Exception as e:
            msg = 'Error reading response from server: ' + str(e)
            logger.debug(msg)
            self.newresponse.text = msg
            return self.newresponse
        self.text = self._links_to_local()
        self.newresponse.text = self.text
        return self.newresponse

class CleanHTMLConverter(RawConverter):
    async def convert(self):
        """Cleans up javascript, styles and excessive formattive format
        """
        warnings.filterwarnings("ignore", category=XMLParsedAsHTMLWarning)
        try:
            self.text = await self.response.text()
        except Exception as e:
            msg = 'Error reading response from server: ' + str(e)
            self.newresponse.text = msg
            return self.newresponse
        self.text = super()._links_to_local()
        try:
            soup = BeautifulSoup(self.text, 'lxml')
        except XMLParsedAsHTMLWarning:
            soup = BeautifulSoup(self.text, 'html5lib')
        for tag in soup.find_all('script', 'iframe', 'frame', 'style'):
            tag.decompose()
        self.text = soup.prettify()  # better formatting
        self.newresponse.text = self.text
        return self.newresponse

class TxtConverter(RawConverter):
    async def convert(self):
        """Only keeps text
        """
        try:
            self.text = await self.response.text()
        except Exception as e:
            msg = 'Error reading response from server: ' + str(e)
            self.newresponse.text = msg
            return self.newresponse
        self.text = super()._links_to_local()

        bs = BeautifulSoup(self.text, 'lxml')
        for tag in bs.find_all('script', 'iframe', 'frame', 'style'):
            tag.decompose()
        self.text = bs.get_text()

        self.newresponse.text = self.text
        self.newresponse.content_type = 'text/plain'
        return self.newresponse


# -*- coding: utf-8 -*-

""" arch-wiki-search (c) Clem Lorteau 2025
License: MIT
"""

import sys
import logging
import asyncio
import traceback
import webbrowser
import urllib.parse
from concurrent.futures import ThreadPoolExecutor

try:
    from __init__ import logger
    from cachingproxy import CachingProxy
    import wikis
except ModuleNotFoundError:
    from arch_wiki_search import logger
    from arch_wiki_search.cachingproxy import CachingProxy
    from arch_wiki_search.wikis import Wikis

class Core:
    """Manages the caching proxy in async context and launches the appropriate browser
    """
    base_url = ''
    search_parm = ''
    current_url = ''
    search_term = ''
    cachingproxy = None
    offline = False
    refresh = False
    debug = False
    _stop = False
    
    async def start(self):
        try:
            await self.cachingproxy.start()
        except Exception as e:
            logger.error(f'Failed to start caching proxy:\n{e}')
            if self.debug:
                print(traceback.format_exc())
            sys.exit(-3)

        await self.cachingproxy.printcachesize()

    async def search(self, search_term = ''):
        url_path = ''
        if search_term != '':
            url_path = self.search_parm + urllib.parse.quote_plus(search_term)
        await self._go(url_path)

    def _openbrowser(self, url):
        try:
            webbrowser.open(url)
        except Exception as e:
            logger.error(f'Failed to start browser: {e}')
            if self.debug:
                print(traceback.format_exc())
        else:
            self.current_url = url
            logger.debug('Calling browser')

    async def _go (self, url_path):
        if (not self.base_url.startswith(('http://', 'https://'))):
            err = f'Unsupported url: {self.base_url}'
            logger.error(err)
            sys.exit(-2)

        dest_url = f'{self.base_url}{url_path}'
        logger.debug(f'Caching and serving {dest_url}')

        #retrieve and if needed cache the requested page before the browser is called
        await self.cachingproxy.fetch(url_path)

        #open browser asynchronously as otherwise the code would be blocking when there's no graphical
        #environment
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor() as executor:
            result = await loop.run_in_executor(executor,
                self._openbrowser, f'http://localhost:{self.cachingproxy.port}/{dest_url}')
        
    async def stop(self):
        self._stop = True
        await self.cachingproxy.stop()
        await self.cachingproxy.printcachesize()

    async def wait(self, secs=2):
        """Sleep and check for stop condition every X seconds
        TODO: proper stop condition
        """
        while not self._stop:
            # if keyboard.is_pressed('q'): self._stop = true #bs needs root
            await asyncio.sleep(secs)

    def __init__(self, knownwikis,
                 base_url=None, search_parm=None,
                 alt_browser='', conv='', wiki='archwiki',
                 offline=False, refresh=False, debug=False, ):
        """base_url (option -u) will override -wiki.url
        search_parm (option -s) will override -wiki.searchstring
        """
        assert knownwikis
        for w in knownwikis:
            if w.name == wiki:
                self.base_url = w.url
                self.search_parm = w.searchstring
                break
            
        if base_url:
            self.base_url = base_url

        # if not self.base_url.endswith('/'):
        #             self.base_url += '/' #so relative links work
        
        if search_parm:
            self.search_parm = search_parm
        self.conv = conv
        self.offline = offline
        self.refresh = refresh
        self.debug = debug

        if self.debug: logger.setLevel(logging.DEBUG)
        else: logger.setLevel(logging.INFO)

        self.cachingproxy = CachingProxy(self.base_url, debug=debug, conv=self.conv)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" arch-wiki-search (c) Clem Lorteau 2025
License: MIT
"""

#TODO: convert html to markdown
#TODO: conv = darkhtml - custom css for dark mode
#TODO: conv = custom css - user supplied css
#TODO: arg to change number of days before cache expiry
#TODO: readme
#TODO: prompt while serving to search other terms

import sys
import asyncio
import argparse

try:
    from __init__ import __name__, __version__, __url__, __newwikirequesturl__, logger
    from exchange import ZIP
    from core import Core
    from wikis import Wikis
except ModuleNotFoundError:
    from arch_wiki_search import __name__, __version__, __url__, __newwikirequesturl__, logger
    from arch_wiki_search.exchange import ZIP
    from arch_wiki_search.core import Core
    from arch_wiki_search.wikis import Wikis

format_blue_underline = '\033[4;34m'
format_yellow = '\x1b[33;20m'
format_bold = '\033[1m'
format_reset = '\033[0m'

async def _main(core, search):
    await core.start()
    try:
        await core.search(search)
        await core.wait()
    except asyncio.CancelledError:
        print('')
        logger.info('Stopping')
    await core.stop()

def main():
    """Load pre-configured base_url/searchstring pairs from yaml file
    """
    knownwikis = Wikis()
    
    parser = argparse.ArgumentParser(
        prog = sys.argv[0],
        description = f'''Read and search Archwiki and other wikis, online or offline, in HTML, markdown or text, on the desktop or the terminal 

Examples:
    {format_yellow}🡪 {format_reset}{sys.argv[0]} \"installation guide\"{format_reset}
    {format_yellow}🡪 {format_reset}{sys.argv[0]} --wiki=wikipedia \"MIT license\"{format_reset}''',
        epilog = f'''Options -u and -s overwrite the corresponding url or searchstring provided by -w
Known wiki names and their url/searchstring pairs are read from a \'{knownwikis.filename}\' file in \'{knownwikis.dirs[0]}\' and \'{knownwikis.dirs[1]}\'
Github: 🌐{format_blue_underline}{__url__}{format_reset}
Request to add new wiki: 🌐{format_blue_underline}{__newwikirequesturl__}{format_reset}''',
        formatter_class = argparse.RawTextHelpFormatter,
    )
    parser.add_argument('-w', '--wiki', default='archwiki',
                         help='Load a known wiki by name (ex: --wiki=wikipedia) [Default: archwiki]',
                         choices=knownwikis.getnames())
    parser.add_argument('-u', '--url', default=None,
                         help='URL of wiki to browse (ex: https://wikipedia.org, https://wiki.freebsd.org)')
    parser.add_argument('-s', '--searchstring', default=None,
                         help='alternative search string (ex: \"/wiki/Special:Search?go=Go&search=\", \"/FrontPage?action=fullsearch&value=\")')
    parser.add_argument('-c', '--conv', default=None,
                        choices=['raw', 'clean', 'txt'],
                        help='''conversion mode:
raw: no conversion (but still remove binaries)
clean: convert to simple html (basic formatting, no styles or scripts)
txt: convert to plain text
[Default: \'raw\' in graphical environment, \'clean\' otherwise]''',)
    parser.add_argument('--offline', '--test', default=False, action='store_true',
                         help='Don\'t try to go online, only use cached copy if it exists')
    parser.add_argument('--refresh', default=False, action='store_true',
                        help='Force going online and refresh the cache')
    parser.add_argument('-v', '--version', default=False, action='store_true',
                        help='Print version number and exit')
    parser.add_argument('-x', '--export', default=False, action='store_true',
                        help='Export cache as .zip file')
    parser.add_argument('-m', '--merge', default=None,
                        help='Import and merge cache from a zip file created with --export') #TODO validate the import
    parser.add_argument('-d', '--debug', default=False, action='store_true')
    parser.add_argument('search', help='string to search (ex: \"installation guide\")', nargs='?',
                        const=None, type=str)
    args = parser.parse_args()
    if (args.version):
        print(__version__)
        sys.exit(0)

    if (not args.search):
        search = ''
    else:
        search = args.search

    core = Core(knownwikis,
                # alt_browser=args.browser,
                conv=args.conv,
                base_url=args.url, 
                search_parm=args.searchstring,
                offline=args.offline,
                refresh=args.refresh,
                debug=args.debug,
                wiki=args.wiki,
                )

    if (args.export):
        if (args.merge):
            logger.critical('--export and --merge can\'t be used together')
            sys.exit(-6)
        ZIP().export(core.cachingproxy.cache_dir)
        sys.exit(0)

    if (args.merge):
        if args.export:
            logger.critical('--export and --merge can\'t be used together')
            zip.exit(-6)
        ZIP().merge(core.cachingproxy.cache_dir, args.merge)
        sys.exit(0)

    try:
        asyncio.run(_main(core, search))
    except KeyboardInterrupt:
        pass #exception CancelledError will be caught in main

if __name__ == '__main__':
    main()